/*
 * hptsb.h  -- An HP-2000 Timeshared BASIC System tape manipulation program.
 *
 * Copyright (c) 2005, Scott G. Taylor
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * SCOTT G. TAYLOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Scott G. Taylor shall not
 * be used in advertising or otherwise to promote the sale, use or other dealings
 * in this Software without prior written authorization from Scott G. Taylor.
 */

#define ENTRY_TYPE_SET_PROT     8
#define ENTRY_TYPE_SET_ASCII    4
#define ENTRY_TYPE_SET_FILE     2
#define ENTRY_TYPE_SET_CSAVE    1
#define ENTRY_TYPE_SET_PROGRAM  0
#define ENTRY_TYPE_PROT(x)      (x & ENTRY_TYPE_SET_PROT)
#define ENTRY_TYPE_ASCII(x)     (x & ENTRY_TYPE_SET_ASCII)
#define ENTRY_TYPE_FILE(x)      (x & ENTRY_TYPE_SET_FILE)
#define ENTRY_TYPE_CSAVE(x)     (x & ENTRY_TYPE_SET_CSAVE)
#define ENTRY_TYPE_PROGRAM(x)   (!( x & ENTRY_TYPE_SET_ASCII) && !(x & ENTRY_TYPE_SET_FILE))

#define	WTOI(buffer)	(((*(buffer) & 0xff) << 8) | (*((buffer) + 1) & 0xff))

#define	DEBUG_HEADERS   3       /*      H       */
#define	DEBUG_SYSTEM    4       /*      S       */
#define	DEBUG_INDCONV   5       /*      I       */
#define	DEBUG_TABLES    6       /*      L       */
#define	DEBUG_FILEREAD  7       /*      R       */
#define	DEBUG_FILES     8       /*      F       */
#define	DEBUG_MEMORY    9       /*      M       */
#define	DEBUG_TAPE      10      /*      P       */
#define	DEBUG_DETOKEN   11      /*      D       */
#define	DEBUG_CONVERT   12      /*      C       */
#define	DEBUG_DETOKEN2  13      /*      E       */
#define	DEBUG_TOKEN     14      /*      T       */

struct line_table {
	uint32_t		len;
	uint8_t			*data;
	struct line_table	*next;
};

struct Program_Entry {
	uint8_t		header[24];
	uint32_t	id;
	uint8_t		HPfilnam[6];
	uint8_t		filnam[7];
	uint32_t	startloc;
	uint32_t	refdate;
	uint32_t	moddate;
	uint32_t	status;
	uint32_t	drumaddr;
	uint32_t	daddr;
	uint32_t	ddev;
	uint32_t	linfo;
	uint32_t	len;
	uint32_t	type;
	uint32_t	filesize;
	uint32_t	symtabstart;
	uint32_t	block_count;
	uint32_t	blockno;
	uint32_t	block_size;
	uint8_t		*block;
	uint32_t	converted_lines;
	uint32_t	converted_bytes;
	struct line_table cline;
};

#define MACH_TYPE_SET_UNKNOWN	0
#define MACH_TYPE_SET_2000A	1100
#define MACH_TYPE_SET_2000B	1200
#define MACH_TYPE_SET_2000C	1300
#define MACH_TYPE_SET_2000E	1500
#define MACH_TYPE_SET_2000CP	2000
#define MACH_TYPE_SET_2000F	3000
#define	MACH_TYPE_SET_2000FP	3500
#define MACH_TYPE_SET_ACCESSA	5000
#define MACH_TYPE_SET_ACCESSB	6000
#define MACH_TYPE_SET_TSC	7000
#define MACH_TYPE_VALID(mtype)	  (mtype == MACH_TYPE_SET_2000A \
				|| mtype == MACH_TYPE_SET_2000B \
			        || mtype == MACH_TYPE_SET_2000C \
				|| mtype == MACH_TYPE_SET_2000E \
			        || mtype == MACH_TYPE_SET_2000CP \
				|| mtype == MACH_TYPE_SET_2000F \
				|| mtype == MACH_TYPE_SET_2000FP \
				|| mtype == MACH_TYPE_SET_ACCESSA \
				|| mtype == MACH_TYPE_SET_ACCESSB \
			        || mtype == MACH_TYPE_SET_TSC)

#define MACH_TYPE_UNKNOWN(mtype)	(mtype == MACH_TYPE_SET_UNKNOWN)
#define MACH_TYPE_2000A(mtype)		(mtype == MACH_TYPE_SET_2000A)
#define MACH_TYPE_2000B(mtype)		(mtype == MACH_TYPE_SET_2000B)
#define MACH_TYPE_2000C(mtype)		(mtype == MACH_TYPE_SET_2000C || mtype == MACH_TYPE_SET_2000CP)
#define MACH_TYPE_2000E(mtype)		(mtype == MACH_TYPE_SET_2000E)
#define MACH_TYPE_2000F(mtype)		(mtype == MACH_TYPE_SET_2000F || mtype == MACH_TYPE_SET_2000FP \
								      || mtype == MACH_TYPE_SET_2000CP \
								      || mtype == MACH_TYPE_SET_TSC)
#define MACH_TYPE_ACCESS(mtype)		(mtype == MACH_TYPE_SET_ACCESSA || mtype == MACH_TYPE_SET_ACCESSB)
#define MACH_TYPE_TSC(mtype)		(mtype == MACH_TYPE_SET_TSC)
#define MACH_TYPE_NODUMP(mtype)		(mtype == MACH_TYPE_2000C || mtype == MACH_TYPE_2000B \
								  || mtype == MACH_TYPE_2000A)
#define MACH_TYPE_DUMP(mtype)		(!MACH_TYPE_NODUMP(mtype))

/*
 * HP System Level Codes - Add to this table as discovered.
 */

/*
 *	Known Hardware Options:
 *		2000/C	001	1,572,864 Byte, Fixed-Head 2774A Drum (16 terminal system only)
 *			002	2,097,152 Byte, Fixed-Head 2774A-003 (16 terminal system only)
 *			003	32-Terminal (Additional 12584 MUX), Plus option 001.
 *			004	32-Terminal (Additional 12584 MUX), Plus option 002.
 *			005	23.905,280 Byte, Moving-Head 2883A Disc.
 *			015	230V, 50Hz option.
 *		2000/F	000
 *			001	23.905,280 Byte, Moving-Head 2883A Disc.
 *			002	1,572,864 Byte, Fixed-Head 2774A Drum
 *			003	2,097,152 Byte, Fixed-Head 2774A-003 (16 terminal system only)
 *			004	1600 BPI tape
 *
 *	Known Software Options:
 *		2000/C	200	High-Speed MUX
 *		2000/F	200	7900A Disc Drive
 *		2000/F	205	2883A Disc Drive
 *		2000/F	210	7900A Disc Drive, High-Speed MUX
 *		2000/F	215	2883A Disc Drive, High-Speed MUX
 *		2000/F	500
 *		2000/F	505
 *		2000/F	510
 *		2000/F	515
 */

#define MACH_TYPE_NAME_UNKNOWN	system_level_name[0]
#define MACH_TYPE_NAME_2000A	system_level_name[1]
#define MACH_TYPE_NAME_2000B	system_level_name[2]
#define MACH_TYPE_NAME_2000E	system_level_name[3]
#define MACH_TYPE_NAME_2000C	system_level_name[4]
#define MACH_TYPE_NAME_2000CP	system_level_name[5]
#define MACH_TYPE_NAME_2000F	system_level_name[6]
#define MACH_TYPE_NAME_2000FP	system_level_name[7]
#define MACH_TYPE_NAME_ACCESSA	system_level_name[8]
#define MACH_TYPE_NAME_ACCESSB	system_level_name[9]

struct equiptable_2000access {
	uint32_t	NIDT;		/* Number of ID tracks				*/
	uint32_t	NDIRT;		/* Number of directory tracks per disc  	*/
	uint32_t	ADTAT[24];	/* ADT address table				*/
	uint32_t	DKTBL[8];	/* Disk Table.  Each word corresponds to a disk */
					/* drive.  If the word is zero, the drive does	*/
					/* not exist.  Bits 15:8 contain the high-	*/
					/* priority select code and 7:0 contain the	*/
					/* unit number.					*/
	uint32_t	DADSL[2];	/* Disc address of system library.		*/
	uint32_t	DLNSL;		/* Length of system library in blocks.		*/
	uint8_t		SYSID[10];	/* Ten-character system ID, space padded.	*/
	uint32_t	MAGSC;		/* High-priority select code for the mag tape.	*/
					/* Bit 15 is set if the console is a terminet.	*/
	uint32_t	NPORT;		/* Two's-complement of the number of ports on	*/
					/* the system.					*/
	uint32_t	YEAR;		/* Sleep/Hibernate year of the century.		*/
	uint32_t	HOUR;		/* Hour of the year.				*/
	uint32_t	TENTHS;		/* Number of 100ms units in the hour minus	*/
					/* 36000.					*/
	uint32_t	HDATE;		/* Hour of the year that the system was last	*/
					/* Hibernated (Same as HOUR).			*/
	uint32_t	SLEPT;		/* 0 = System has been slept, -1 = has not.	*/
					/* This word ensures that the system may not be	*/
					/* reloaded if it has not been slept.  For 	*/
					/* tapes, this is always set to 1.		*/
};

struct equiptable_2000C {
	uint32_t	IDEC[9];	/* ID table headers				*/
	uint32_t	NIDT;		/* Number of ID tracks				*/
	uint32_t	ADLOC;		/* Drum address of ADT				*/
	uint32_t	ADLEN;		/* Length of ADT in -words			*/
	uint32_t	NDIRT;		/* Number of directory tracks per disc  	*/
	uint32_t	DADLC[8];	/* Disc ADT drum addresses			*/
	uint32_t	DADLN[8];	/* Disc ADT lengths in -words			*/
	uint32_t	DTBL[4];	/* There is one word in this area for each of	*/
					/* the 4 drums.  When the word is zero, the	*/
					/* particular drum does not exist.  Otherwise,	*/
					/* bits 7:6 contain the drum prefix and bits	*/
					/* 5:0 the high priority select code.  The	*/
					/* prefix is used by the drum driver as the	*/
					/* high order 2 bits of the 8-bit track address.*/
	uint32_t	DKTBL[8];	/* Disk Table.  Each word corresponds to a disk */
					/* drive.  If the word is zero, the drive does	*/
					/* not exist.  Bits 15:8 contain the high-	*/
					/* priority select code and 7:0 contain the	*/
					/* unit number.					*/
	uint32_t	TRAX[16];	/* This is a table of which drum tracks are	*/
					/* physically available to the system.  	*/
					/* Locations 151-154 (words 0-3) correspond	*/
					/* to drum 0, 155-156 (words 4-7) to drum 1,	*/
					/* etc.  Track 0 of drum 0 is represented by	*/
					/* bit 0 of 151 (word 0), track 1 of drum 0 is	*/
					/* represented by bit 1 of 151 (word 0), etc.	*/
					/* A bit is 0 when the track is available,	*/
					/* 1 when unavailable.				*/
					/* The TRAX table is changed only by the	*/
					/* following commands:				*/
					/* DRUM   - causes all tracks of the specified	*/
					/*          drum to be made available.		*/
					/* LOCK   - all specified tracks are made	*/
					/*          unavailable.			*/
					/* UNLOCK - all specified tracks are made	*/
					/*          available.				*/
	uint8_t		SYSID[10];	/* Ten-character system ID, space padded.	*/
	uint32_t	MAGSC;		/* High-priority select code for the mag tape.	*/
					/* Bit 15 is set if the tape unit is a 7970.	*/
	uint32_t	NPORT;		/* Two's-complement of the number of ports on	*/
					/* the system.					*/
	uint32_t	YEAR;		/* Sleep/Hibernate year of the century.		*/
	uint32_t	HOUR;		/* Hour of the year.				*/
	uint32_t	TENTHS;		/* Number of 100ms units in the hour minus	*/
					/* 36000.					*/
	uint32_t	HDATE;		/* Hour of the year that the system was last	*/
					/* Hibernated (Same as HOUR).			*/
	uint32_t	SLEPT;		/* 0 = System has been slept, -1 = has not.	*/
					/* This word ensures that the system may not be	*/
					/* reloaded if it has not been slept.  For 	*/
					/* tapes, this is always set to 1.		*/
};

struct equiptable_2000F {
	uint32_t	IDEC[12];	/* ID table headers.				*/
					/* Each 4-word entry consists of:		*/
					/*   1) First ID on the track			*/
					/*   2) Disc					*/
					/*   3)   address				*/
					/*   4) Length in -words			*/
	uint32_t	NIDT;		/* Number of ID tracks				*/
	uint32_t	NDIRT;		/* Number of directory tracks per disc  	*/
	uint32_t	ADTAT[24];	/* ADT Disc address table.			*/
					/* Each 3-word entry consists of:		*/
					/*   1) Disc					*/
					/*   2)   address				*/
					/*   3) Length in -words			*/
	uint32_t	ADTLN;		/* Length of Disc 0. (ADTAT+2)			*/
	uint32_t	DKTBL[8];	/* Disk Table.  Each word corresponds to a disk */
					/* drive.  If the word is zero, the drive does	*/
					/* not exist.  Bits 15:8 contain the high-	*/
					/* priority select code and 7:0 contain the	*/
					/* unit number.					*/
	uint32_t	DADSL[2];	/* Disc address of system library.		*/
	uint32_t	DLNSL;		/* Disc length of system library.		*/
	uint8_t		SYSID[10];	/* Ten-character system ID, space padded.	*/
	uint32_t	MAGSC;		/* High-priority select code for the mag tape.	*/
					/* Bit 15 is set if the tape unit is a 7970.	*/
	uint32_t	NPORT;		/* Two's-complement of the number of ports on	*/
					/* the system.					*/
	uint32_t	YEAR;		/* Sleep/Hibernate year of the century.		*/
	uint32_t	HOUR;		/* Hour of the year.				*/
	uint32_t	TENTHS;		/* Number of 100ms units in the hour minus	*/
					/* 36000.					*/
	uint32_t	HDATE;		/* Hour of the year that the system was last	*/
					/* Hibernated (Same as HOUR).			*/
	uint32_t	SLEPT;		/* 0 = System has been slept, -1 = has not.	*/
					/* This word ensures that the system may not be	*/
					/* reloaded if it has not been slept.  For 	*/
					/* tapes, this is always set to 1.		*/
};

struct equiptable_2000E {
	struct {
		uint32_t	id;
		uint32_t	filnam[3];
		uint32_t	address;
	} DIREC [4];
	uint32_t	IDLEN;
	uint32_t	IDTTA;
	uint32_t	IDTAT[4];
	uint32_t	IDTRL;
	uint32_t	IDTLT[4];
	uint32_t	ADLOC;
	uint32_t	ADLEN;
	uint32_t	SDT;
	/*uint32_t	ADTBL;*/
	uint32_t	NPORT;
	uint32_t	PHR;
	uint32_t	SLEPT;
};

/*
 * DIREC Table Format
 *
 *   NOTE: 2000/ACCESS system tapes have 40 words (80 bytes) preceeding the
 *	   bytes described below.  The described bytes are transcribed from
 *	   the HP documentation, but the additional 80 bytes are so-far
 *	   undocumented.
 *
 *      Word	Content
 *	 0	-length in words of first directory track
 *      1-4	Same as first four words of first Directory Entry track
 *      5-6	Disc address (word 5 unused on 2000A systems).
 *      7-13	Same as 0-6 but applied to 2nd directory track.
 *     14-20	Same as 0-6 but applied to 3rd directory track.
 *     21-27	Same as 0-6 but applied to 4th directory track.
 *
 *	   (The following words are for 2000/ACCESS systems)
 *     28-33	Same as 0-6 but applied to 5th directory track.
 *		.
 *		.
 *		.
 *   553-559	Same as 0-6 but applied to 80th directory track.
 *
 *	A disc address of 0 implies that there is no such directory track.
 *	When word 0 is 0, words 1-4 are meaningless.
 *
 *	EARLY 2000 systems;
 *		The disc address of a directory is always sector 0 of a track.
 *		each directory track may contain as many as 5440 words = 85
 *		sectors = 680 directory entries.  Directory tracks are
 *		allocated as follows:
 *			a.)  When the system is initial loaded, or when it is
 *			     loaded from mag. tape, the number of directory
 *			     tracks allocated is the maximum of the original
 *			     number of tracks and the number of discs;
 *			b.)  When the DISC command is used to add a disc, a new
 *			     directory track is allocated unless this would cause
 *			     there to be more directory tracks than discs.
 *
 *	2000/ACCESS systems;
 *		When generating a system or reloading from mag tape, the system
 *		operator has the opportunity to specify the number of directory
 *		tracks per disc, in the range of 1-10, which is saved in NDIRT.
 *		The total number of directory tracks is this number times the
 *		number of discs on the system.  Each directory track may contain
 *		as many as 8184 words = 682 directory entries.
 *
 */

struct DIRECtable_2000access {
	struct {
		uint32_t	len;
		uint32_t	first_directory_userid;
		uint8_t		first_directory_name[6];
		uint32_t	disc_address_1;
		uint32_t	disc_address_2;
	} track[80];
};

uint8_t		* nDtab, * nDtabp;

struct DIRECTORYtable {
	uint32_t	id;
	uint8_t *	name;
	uint32_t	start;
	uint32_t	refdate;
	uint32_t	moddate;
	uint32_t	data; 		/* 2000C/2000F = drumaddr, ACCESS = status */
	uint32_t	daddr;
	uint32_t	ddev;
	uint32_t	linfo;
	uint32_t	len;
};

struct NONSHAREDEVICEtable_2000access {
	uint32_t	ndevices;
	struct {
		uint32_t	data1;
		uint32_t	data2;
		uint32_t	data3;
		uint32_t	data4;
		uint32_t	data5;
	} entry[64];
};

struct SystemSegmentTable {
	uint32_t		count;
	struct {
		uint32_t	len;
		uint32_t	start;
		uint32_t	addr[2];
		uint8_t	*	block;
	} rec[50];
};

struct SystemLibraryTable {
	uint32_t		count;
	struct {
		uint32_t	len;
		uint32_t	start;
		uint32_t	addr[2];
		uint8_t	*	block;
	} rec[50];
};


#define	HPyear(HPdate)	((HPdate >> 9) + 1900)
#define HPday(HPdate)	(HPdate & 0x1ff)

/*
 *			BASIC Internal Representation
 *	BASIC statements are represented internally by the sequence
 *	number followed by the length in workds (including the sequence
 *	number and length words) followed by the statement body.  The
 *	statement body is composed almost entirely of operator-operand
 *	pairs which occupy from one to three words each.  Null operands
 *	and operators are used when necessary to maintain the
 *	operator-operand correspondence.  The operator resides in bits
 *	14-9 of a word; the operand uses bit 15, bits 8-0, and sometimes
 *	whole additional words immediate following.
 *
 * 	Variable Operands:
 *	-----------------
 *
 *	---------------------------------
 *	: 0 : Operator :               0:	Null Operand
 *	---------------------------------
 *
 *	---------------------------------
 *	: 0 : Operator : Name :    0    :	String Variable
 *	---------------------------------
 *
 *	---------------------------------
 *	: 0 : Operator : Name :  1-3    :	Array Variable
 *	---------------------------------
 *
 *	---------------------------------
 *	: 0 : Operator : Name : 4-16(8) :	Simple Variable
 *	---------------------------------
 *
 *	---------------------------------
 *	: 0 : Operator : Name : 17(8)   :	Function Variable
 *	---------------------------------
 *
 *	Bits 8-0 are generally divided into two fields as follows:
 *	a name field (bits 8-4) and a type field (bits 3-0).  The name
 *	field holds a value between 1 and 32(8) corresponding to A-Z (for
 *	functions, corresponding to FNA through FNZ).  A type of 0
 *	identifies a string variable (e.g. 3,0 represents C$).  Types 1 and
 *	2 identify array variables of dimensionality one and two
 *	respectively (e.g. 4,2 represents D[*,*] while type 3 identifies
 *	an array variable whose dimensionality cannot be determined by its
 *	immediate context.  Type 4 identifies a simple variable with no
 *	digit (e.g. 1,4 represents A) while types 5-16(8) identify
 *	simple variables whose names include the digit 0-9(10)
 *	respectively (e.g. 6,7 represents F2).  Type 17(8) identifies a
 *	programmer-defined function (e.g. 32(8), 17(8) represents FNZ).
 *	Name field values of 33(8) thru 36(8) correspond to extended
 *	string representations (e.g. A0$ thru Z1$).
 *
 *	Constant Operands:
 *	-----------------
 *
 *	---------------------------------
 *	: 1 : Operator : Name : 4-16(8) :	Parameter
 *	---------------------------------
 *
 *	---------------------------------
 *	: 1 : Operator : Name : 17(8)   :	Pre-defined Function
 *	---------------------------------
 *
 *	---------------------------------
 *	: 1 : Operator :        3       :	Formal Dimension/
 *	---------------------------------
 *	:        Binary Integer         :	Branch Address
 *	---------------------------------	  List
 *	:                               :
 *	:                               :
 *	---------------------------------
 *	:        Binary Integer         :
 *	---------------------------------
 *
 *	---------------------------------
 *	: 1 : Operator :        0       :	Numerical Constant
 *	---------------------------------
 *	:        High Mantissa          :
 *	---------------------------------
 *	: Low Mant     : Exponent       :
 *	---------------------------------
 *
 *	---------------------------------
 *	: 0 : 1 (") :  0255(10)         :
 *	---------------------------------	String Constant
 *	:                               :
 *	:                               :
 *	---------------------------------	String Constant
 *	: Character    : Character      :
 *	---------------------------------
 *
 *	A parameter (which can only appear inside a <DEF statement>)
 *	differs from a simple variable only in that bit 15 is set.  The
 *	name of a predefined function may range, in the standard system,
 *	from 0 to 31(8) or 33(8) to 37(8) (CTL to SYS or ZER to TRN).  A
 *	flagged (bit 15 set) operand of 3 identifies either a formal
 *	dimension in a <DIM statement> or <COM statement> (value in
 *	following word) or a branch address list (one or more statement
 *	sequence numbers in the following words).  A flagged operand of
 *	0 indicates that the following two words hold a floating-point
 *	constant (all numerical constants within a program are so
 *	represented).  The operator with internal code 1 is ", which
 *	signals the start of a string constant.  The operand portion of
 *	the word has a value from 0 to 255(10), indicating the number of
 *	characters in the constant.  The string follows, two characters
 *	per word, and the closing " is not explicitly represented
 *	internally.
 */
